#include "psi3d.h"


//
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
//


#define distance2d(A, B)       (sqrt( ((A)[X]-(B)[X])*((A)[X]-(B)[X])+((A)[Y]-(B)[Y])*((A)[Y]-(B)[Y]) ))


cBone::cBone()
{
  mName      = NULL;
  mTimer     = 0;
  mChildren  = NULL;
  mNext      = NULL;
}


BOOLEAN cBone::Hit(cSphere other)
{
  cSphere sphere;
  cVector v[2];
  v[0] = mUJoint;
  v[1] = mLJoint;
  sphere.Set(v, 2);
  return sphere.Hit(other);
}


INTEG cBone::FindBone(char *name, INTEG n, cBone *array)

{

  INTEG i;

  for (i=0; i<n; i++)

  {

    if (array[i].mName != NULL)

    if (strcmp(name, array[i].mName) == 0) return i;

  };

  return -1;

}



void cBone::Scale(SCALAR f)
{
  mUBase.Mul(f);
  mLBase.Mul(f);
  mUJoint.Mul(f);
  mLJoint.Mul(f);
  mOUJoint.Mul(f);
  mOLJoint.Mul(f);
}


void cBone::Animate(SCALAR delta, INTEG n, cVector *keyFrames)

{

  mTimer += delta;

  mAnim = mAnim.GetInterpolated(mTimer, n, keyFrames);

}





void cBone::ToWorld(cMatrix super, cVector length)

{

  mMatrix.Set();

  mMatrix.RotXYZ(mCenter.mV[X], mCenter.mV[Y], mCenter.mV[Z]);

  mMatrix.RotXYZ(mAnim.mV[X], mAnim.mV[Y], mAnim.mV[Z]);

  mMatrix.MovRel(length.mV[X], length.mV[Y], length.mV[Z]);

  mMatrix.Mul(super);

  

  mOUJoint = mUJoint;
  mUJoint.Set(0, 0, 0);

  mMatrix.Apply(mUJoint);



  mOLJoint = mLJoint;
  mLJoint = mLength;

  mMatrix.Apply(mLJoint);



  cBone *bone = mChildren;

  while (bone)

  {

    bone->ToWorld(mMatrix, mLength);

    bone = bone->mNext;

  };

}





void cBone::BaseToWorld(cMatrix super, cVector length)

{

  mMatrix.Set();

//  mMatrix.RotXYZ(mCenter.mV[X], mCenter.mV[Y], mCenter.mV[Z]);

  mMatrix.MovRel(length.mV[X], length.mV[Y], length.mV[Z]);

  mMatrix.Mul(super);

  

  mUBase.Set(0, 0, 0);

  mMatrix.Apply(mUBase);



  mLBase = mLength;

  mMatrix.Apply(mLBase);



  cBone *bone = mChildren;

  while (bone)

  {

    bone->BaseToWorld(mMatrix, mLength);

    bone = bone->mNext;

  };

}





void cBone::ToWorld(cMatrix super, cMatrix rsuper, cVector length)

{

  mMatrix.Set();

  mMatrix.RotXYZ(mCenter.mV[X], mCenter.mV[Y], mCenter.mV[Z]);

  mMatrix.RotXYZ(mAnim.mV[X], mAnim.mV[Y], mAnim.mV[Z]);

  mMatrix.MovRel(length.mV[X], length.mV[Y], length.mV[Z]);

  mMatrix.Mul(super);

  

  mOUJoint = mUJoint;
  mUJoint.Set(0, 0, 0);

  mMatrix.Apply(mUJoint);



  mOLJoint = mLJoint;
  mLJoint = mLength;

  mMatrix.Apply(mLJoint);



  cVector r;

  r.Add(mAnim);



  mLimb.Set();

  mLimb.RotXYZ(mCenter.mV[X], mCenter.mV[Y], mCenter.mV[Z]);

  mLimb.RotXYZ(r.mV[X], r.mV[Y], r.mV[Z]);

  mLimb.Mul(rsuper);



  cBone *bone = mChildren;

  while (bone)

  {

    bone->ToWorld(mMatrix, mLimb, mLength);

    bone = bone->mNext;

  };

}





void cBone::Draw(cMatrix camera)

{

  cMatrix c;

  c = mMatrix;

  c.Mul(camera);

  glLoadMatrixf(c.mM);



  glDisable(GL_TEXTURE_2D);

  glBegin(GL_LINES);

    glColor4f(0.0, 1.0, 0.0, 1.0);

    glVertex3f(0.0, 0.0, 0.0);

    glColor4f(0.0, 0.0, 1.0, 1.0);

    glVertex3fv(mLength.mV);

  glEnd();



  cBone *bone = mChildren;

  while (bone)

  {

    bone->Draw(camera);

    bone = bone->mNext;

  };

};





//----------------------------->





cAniBone::cAniBone()

{

  mName      = NULL;

  mBone      = -1;

  mNKFrames  = 0;

  mKFrames   = NULL;

}





void cAniBone::Load(FILE *f)

{

  char buf[80];

  INTEG i, n;

  SCALAR a, b, c, d;

  fscanf(f, "%s %i\n", buf, &n);

//fprintf(errlog, "%s %i\n", buf, n);
//fflush(errlog);
  mName = new char[strlen(buf)+1];

  strcpy(mName, buf);

  mNKFrames = n;

  mKFrames = new cVector[mNKFrames];

  for (i=0; i<n; i++)

  {

    fscanf(f, "%f, %f %f %f\n", &d, &a, &b, &c);

    mKFrames[i].Set(a, b, c, d);

  };

}





//----------------------------->





cAniBody::cAniBody()

{

  mName         = NULL;

  mBody         = -1;

  mNAniBones    = 0;

  mAniBones     = NULL;

}





void cAniBody::Load(FILE *f)

{

  char buf[80];

  INTEG i, n;

  fscanf(f, "%s %i\n", buf, &n);

//fprintf(errlog, "%s %i\n", buf, n);
//fflush(errlog);
  mName = new char[strlen(buf)+1];

  strcpy(mName, buf);

  mNAniBones = n;

  mAniBones = new cAniBone[mNAniBones];

  for (i=0; i<n; i++)

  {

    mAniBones[i].Load(f);

  };

}





void cAniBody::Assign(INTEG n, cBone *array)

{

  int i;

  for (i=0; i<mNAniBones; i++)

  {

    mAniBones[i].mBone = array[0].FindBone(mAniBones[i].mName, n, array);

  };

}





//----------------------------->





void cAnime::Load(FILE *f)

{

  char buf[80];

  INTEG i, n;

  FILE *xf;

  fscanf(f, "%i\n", &n);
//fprintf(errlog, "%i\n", n);
//fflush(errlog);
  mNAnimations = n;

  mAnimations = new cAniBody[mNAnimations];

  for (i=0; i<n; i++)

  {

    fscanf(f, "%s\n", buf);

    xf = fopen(buf, "r");

    mAnimations[i].Load(xf);

    fclose(xf);

  };

}





void cAnime::Assign(INTEG n, cBone *array)

{

  int i;

  for (i=0; i<mNAnimations; i++)

  {

    mAnimations[i].Assign(n, array);

  };

}





//----------------------------->





cBody::cBody()

{

  mNBones  = 0;

  mBones   = NULL;

}





BOOLEAN cBody::Hit(cSphere other)

{

  INTEG i;

  for (i=0; i<mNBones; i++)

  {

    if (mBones[i].Hit(other)) return 1;

  };

  return 0;

}





cBone *cBody::LoadSub(FILE *f)

{

  char buf[80];

  INTEG i, n, children;

  SCALAR a, b, c;

  fscanf(f, "%s %i, %f %f %f\n", buf, &children, &a, &b, &c);

  n = mBones[0].FindBone(buf, mNBones, mBones);

//fprintf(errlog, "<%s> <%i>\n", buf, n);
//fflush(errlog);
  if (n < 0) return (cBone*) NULL;

  mBones[n].mLength.Set(a, b, c);

  cBone *bone = (cBone*) &mBones[n];

  for (i=0; i<children; i++)

  {

    if (i == 0) {

      bone->mChildren = LoadSub(f);

      bone = bone->mChildren;

    } else {

      bone->mNext = LoadSub(f);

      bone = bone->mNext;

    };

  };

  return (cBone*) &mBones[n];

}





void cBody::Load(FILE *f)

{

  char buf[80];

  INTEG i, n;

  SCALAR a, b, c;

  fscanf(f, "%i\n", &n);

  mNBones = 0;

  mBones = new cBone[n];

  for (i=0; i<n; i++)

  {
    fscanf(f, "%s %f %f %f\n", buf, &a, &b, &c);
    mBones[mNBones].mCenter.Set(a, b, c);
    mBones[mNBones].mName = new char[strlen(buf)+1];
//fprintf(errlog, "<%s>", buf);
    strcpy(mBones[mNBones].mName, buf);
//fprintf(errlog, "<%s>\n", mBones[mNBones].mName);
//fflush(errlog);
    mNBones++;

  };

//fprintf(errlog, "loadsub...\n");
//fflush(errlog);
  LoadSub(f);
}


void cBody::SaveSub(FILE *f, cBone *bone)

{

  INTEG i, children = 0;

  cBone *next = bone->mChildren;

  cVector l = bone->mLength;

  while (next) // count amount of children

  {

     children++;

     next = next->mNext;

  };

  fprintf(f, "%s %i, %f %f %f\n", bone->mName, children, l.mV[X], l.mV[Y], l.mV[Z]);

  next = bone->mChildren;

  while (next)

  {

     SaveSub(f, next);

     next = next->mNext;

  };

}





void cBody::Save(FILE *f)

{

  INTEG i;

  SCALAR a, b, c;

  fprintf(f, "%i\n", mNBones);

  for (i=0; i<mNBones; i++)

  {

    a = mBones[i].mCenter.mV[X];

    b = mBones[i].mCenter.mV[Y];

    c = mBones[i].mCenter.mV[Z];

    fprintf(f, "%s %f %f %f\n", mBones[i].mName, a, b, c);

  };

  SaveSub(f, mBones);

}





void cBody::Scale(SCALAR f)

{

  INTEG i;

  for (i=0; i<mNBones; i++)

  {

    mBones[i].mLength.Mul(f);

  };

}





INTEG cBody::Select(cVector pos)

{

  INTEG i, a;

  double e, d;

  a = -1;

  d = 100000000.0;

  for (i=0; i<mNBones; i++)

  {

    e = distance2d(pos.mV, mBones[i].mLJoint.mV);

    if (e<d) {

      d = e;

      a = i;

    };

  };

  return a;

}





void cBody::Drag(cVector pos)

{

  INTEG a = Select(pos);

  if (a>=0) {

    cVector offset = pos;

    offset.Sub(mBones[a].mLJoint);

    mBones[a].mLength.Add(offset);

  };

}



