#include "psi3d.h"


#ifdef USE_YIFF2

extern "C" {
#include <Y2/Y.h>
#include <Y2/Ylib.h>
}

#endif



cSound::cSound()
{
  mSystemHandle  = NULL;
  mNRegistered   = 0;
  mNPlaying      = 0;
  mRate          = 22050;
  mBits          = 16;
  mChannels      = 2;
  mMaster        = 1;
  mVolumeL       = 1;
  mVolumeR       = 1;
}


cSound::~cSound()
{
  if (mSystemHandle != NULL) KillSystem();
}


void cSound::SetMasterVolume(SCALAR v)
{
  mMaster = v;
}


void cSound::SetVolume(SCALAR l, SCALAR r)
{
  mVolumeL = l;
  mVolumeR = r;
}


void cSound::SetChannels(INTEG n)
{
  mChannels = n;
#ifdef USE_YIFF2
  YSetAudioModeValues((YConnection*) mSystemHandle, mBits, mChannels, mRate, 0, 1, 2, 256);
#endif
}


void cSound::SetRate(INTEG n)
{
  mRate = n;
#ifdef USE_YIFF2
  YSetAudioModeValues((YConnection*) mSystemHandle, mBits, mChannels, mRate, 0, 1, 2, 256);
#endif
}


void cSound::SetBits(INTEG n)
{
  mBits = n;
#ifdef USE_YIFF2
  YSetAudioModeValues((YConnection*) mSystemHandle, mBits, mChannels, mRate, 0, 1, 2, 256);
#endif
}


void cSound::InitSystem(char *address, WORD port)
{
  char arg[256];
  sprintf(arg, "%s:%i", address, port);
#ifdef USE_YIFF2
  mSystemHandle = (void*) YOpenConnection(NULL, arg);
  if (mSystemHandle == NULL) report("Could Not Connect To Y2 Server!\n") else report("Connected to Y2 Server!");
#endif
}


void cSound::UpdateSystem()
{
#ifdef USE_YIFF2
  YEvent event;
  if (YGetNextEvent((YConnection*) mSystemHandle, &event, 0) > 0) {
    switch(event.type) {
      case YSoundObjectKill:
      break;
      case YDisconnect:
      break;
      case YShutdown:
      break;
    };
  };
#endif
}


void cSound::KillSystem()
{
#ifdef USE_YIFF2
  YCloseConnection((YConnection*) mSystemHandle, 1);
#endif
}


DWORD cSound::RegisterSound(char *fname, INTEG priority)
{
  mFNames[mNRegistered] = new char[strlen(fname)];
  strcpy(mFNames[mNRegistered], fname);
  mPriority[mNRegistered] = priority;    // later for sound scheduling...
  mHandles[mNRegistered] = mNRegistered; // later for use with other soundlib/server maybe...
  mNRegistered++;
  return (mNRegistered-1);
}


// note: think a bit about faking 3d sound via panning...
// anyway this function returns a handle to the *playing* instance.

DWORD cSound::InitSound(DWORD handle, INTEG begin, INTEG repeat, SCALAR l, SCALAR r)
{
#ifdef USE_YIFF2
  YEventSoundPlay value;

  value.flags = YPlayValuesFlagPosition |
                YPlayValuesFlagTotalRepeats |
                YPlayValuesFlagVolume |
                YPlayValuesFlagSampleRate;

  value.position       = begin;
  value.total_repeats  = repeat;
  value.left_volume    = mMaster * l; // scale by Master-Volume
  value.right_volume   = mMaster * r;
  value.sample_rate    = 0;

  return YStartPlaySoundObject((YConnection*) mSystemHandle, mFNames[handle], &value);
#endif
}


// pass in the *playing* handle as given by InitSound().

void cSound::KillSound(DWORD handle)
{
#ifdef USE_YIFF2
  YDestroyPlaySoundObject((YConnection*) mSystemHandle, handle);
#endif
}



