#include <fcntl.h>
#include <unistd.h>
#include <signal.h>



class cInput { // don't change!!
  public:
  int buttons, x, y;
};


const unsigned int JS_RETURN = sizeof(cInput);


char        *js0name = "/dev/js0",   *js1name = "/dev/js1";
int                       js0file,                 js1file;
int                     js0status,               js1status;
int                         js0xm,                   js0ym;
int                         js1xm,                   js1ym;
cInput                      js0,                     js1;
cInput                      kb0,                     kb1;
cInput                      ms0,                     ms1;


//
// LinWarrior 3D (Currently this is more a Test-Driver rather than the final game source).
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
// aka hackCraft
//


#define USE_GL
#define NO 0
#define YES (!NO)

#include <psi3d/psi3d.h>   // ;) almost allmighty cosmic psi3d library ;)

#include "config.h"        // "hardcoded" configuration file...change it!
#include "objects.h"       // objects and world database
#include "state.h"         // general purpose state engine


#ifdef __linux
#include <GL/xmesa.h>
#endif




void Fatal(char *errorString)
{

#ifdef USE_3DFX

  XMesaSetFXmode(XMESA_FX_WINDOW);

#endif

  if (errlog != NULL) {
    fprintf(errlog, "%s\n", errorString);
    fflush(errlog);
  } else {
    fprintf(stderr, "%s\n", errorString);
    fflush(stderr);
  };

  exit(1);
}




BOOLEAN LoadTgaTex(char *prefix, char *fname, BOOLEAN transparency, BOOLEAN mipmap, WORD &TexID)
{
  fprintf(errlog, "%s\n", fname);
  fflush(errlog);
  cBmp *b = new cBmp;

  cVector pink;
  char pf[1024];
  sprintf(pf, "%s%s", prefix, fname);
  if (b->LoadTga(pf)) return 1;

  if (transparency) {
    pink.Set(255, 0, 255);
    b->MakeAlpha(pink);        // make pink pixels transparent
    b->Mul(W, (SCALAR) 0.7);            // make more transparent
  };

  return b->glRegister(mipmap, TexID);
}







cConio3d             *con[2], *cDummy;

cViewer              *mainView;

cPose                *mainCamera;

cTimer               *mainFramer;

cParticle            *tmparticle;

cMech                *mech0,
                     *mech1,
                     *mech2;

cWeapon              *weapon0,
                     *weapon1;

cTree                *tree;

cLandscape           *land;

cWorld               *planet;


INTEG                 width       = 0;
INTEG                 height      = 0;



BOOLEAN               rButton     = NO;
BOOLEAN               mButton     = NO;
BOOLEAN               lButton     = NO;

SCALAR                xrSpeed     = 0;
SCALAR                yrSpeed     = 0;

SCALAR                mxpos       = 0;
SCALAR                mypos       = 0;


BOOLEAN               console0          = START_WITH_CONSOLE_0;
BOOLEAN               console1          = START_WITH_CONSOLE_1;
BOOLEAN               paused            = NO;
BOOLEAN               external_camera   = START_WITH_EXTERNAL_CAM;
BOOLEAN               js_disabled       = JOYSTICK_CONTROL_SUX;
BOOLEAN               wireframe         = START_WITH_WIREFRAME;
BOOLEAN               fullscreen        = 0; // <- don't change!!

INTEG                 NTex              = 0;
WORD                  TexID[100];
WORD                  LandID[10];
WORD                  WaterID[10];

cSound               *sound;
DWORD                 SndID[100];

INTEG                 NSky              = 0;
INTEG                 SkyID             = 0;

INTEG                 NGnd              = 0;
INTEG                 GndID             = 0;

INTEG                 NTre              = 0;
INTEG                 TreID             = 0;

INTEG                 NMsc              = 0;
INTEG                 MscID             = 0;



void Init()
{
  INTEG i, j;
  float Black[3] = { 0.0, 0.0, 0.1 };
  float lightColor[3] = { 0.5, 0.2, 1.0 };
  float fogColor[4] = { 1.0, 1.0, 1.0, 1.0 };
  float fogDensity = 0.001;

//  glDisable(GL_DITHER);

  glEnable(GL_BLEND);
  glEnable(GL_DEPTH_TEST);

  glClearColor(0.0, 0.125, 0.0, 0.0);
  glClearDepth(1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


//  glShadeModel(GL_FLAT);
  glShadeModel(GL_SMOOTH);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDepthMask(GL_TRUE);
//  glDepthFunc(GL_LEQUAL);


  glFogi(GL_FOG_MODE, GL_EXP);
  glFogfv(GL_FOG_COLOR, fogColor);
  glFogf(GL_FOG_DENSITY, fogDensity);
  glHint(GL_FOG_HINT,GL_FASTEST);

  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
//  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

/*
  if (
  ) Fatal("Can't Load TGA Ground Textures");
*/

  NTex = 0;
  GndID = NTex;
  if (
    LoadTgaTex(DATA_DIRECTORY, "textures/ground1.tga",   0,1, TexID[NTex++]) |
    LoadTgaTex(DATA_DIRECTORY, "textures/ground0.tga",   1,1, TexID[NTex++])
  ) Fatal("Can't Load TGA Ground Textures");

  SkyID = NTex;
  if (
    LoadTgaTex(DATA_DIRECTORY, "textures/clouds0.tga",   1,1, TexID[NTex++]) |
    LoadTgaTex(DATA_DIRECTORY, "textures/clouds1.tga",   1,1, TexID[NTex++])
  ) Fatal("Can't Load TGA Sky Textures");


  TreID = NTex;
  if (
    LoadTgaTex(DATA_DIRECTORY, "textures/trees/00.tga",  1,1, TexID[NTex++]) |
    LoadTgaTex(DATA_DIRECTORY, "textures/trees/01.tga",  1,1, TexID[NTex++]) |
    LoadTgaTex(DATA_DIRECTORY, "textures/trees/02.tga",  1,1, TexID[NTex++]) |
    LoadTgaTex(DATA_DIRECTORY, "textures/trees/03.tga",  1,1, TexID[NTex++])
  ) Fatal("Can't Load TGA Tree Textures");


  MscID = NTex;
  if (
    LoadTgaTex(DATA_DIRECTORY, "textures/house0.tga",    0,1, TexID[NTex++]) |
    LoadTgaTex(DATA_DIRECTORY, "textures/font0.tga",     1,0, TexID[NTex++]) |
    LoadTgaTex(DATA_DIRECTORY, "textures/flare0.tga",    1,1, TexID[NTex++]) |
    LoadTgaTex(DATA_DIRECTORY, "textures/tile0.tga",     0,1, TexID[NTex++]) |
    LoadTgaTex(DATA_DIRECTORY, "textures/water0.tga",    1,1, TexID[NTex++])
  ) Fatal("Can't Load TGA Misc. Textures");


  con[0] = new cConio3d;
  con[0]->Resize(16, 8, 0.02);
  con[0]->SetFont(TexID[MscID+1]);
  con[0]->Insert("--------------\n");
  con[0]->Insert("Lin Warrior 3D\n");
  con[0]->Insert("--------------\n");
  con[0]->Insert("              \n");
  con[0]->Insert("3D Console 0  \n");
  con[0]->Insert("              \n");
  con[0]->Insert("State: ACTIVE \n");
  con[0]->mPose.mPos.mV[X] = +0.29;
  con[0]->mPose.mPos.mV[Y] = +0.27;
  con[0]->mPose.mRot.mV[X] = +25;
  con[0]->mPose.mRot.mV[Y] = -25;
  con[0]->mCenter.Set(-8.0*0.02, 4*0.02, 0);

  con[1] = new cConio3d;
  con[1]->Resize(16, 8, 0.02);
  con[1]->SetFont(TexID[MscID+1]);
  con[1]->Insert("--------------\n");
  con[1]->Insert("Lin Warrior 3D\n");
  con[1]->Insert("--------------\n");
  con[1]->Insert("              \n");
  con[1]->Insert("3D Console 1  \n");
  con[1]->Insert("              \n");
  con[1]->Insert("State:INACTIVE\n");
  con[1]->mPose.mPos.mV[X] = -0.29;
  con[1]->mPose.mPos.mV[Y] = +0.27;
  con[1]->mPose.mRot.mV[X] = +25;
  con[1]->mPose.mRot.mV[Y] = +25;
  con[1]->mCenter.Set(-8.0*0.02, 4*0.02, 0);

  cDummy = con[0];

  mainView        = new cViewer;
  mainFramer      = new cTimer;
  mainCamera      = new cPose;
  planet          = new cWorld;
  planet->mDome   = new cDome;

  mainCamera->mPos.Set(0, 1.85, 0);

  FILE *file;


  INTEG a;
  INTEG b;
  cVector t;
  char buf[1024];
  sprintf(buf, "%s%s", DATA_DIRECTORY, "textures/hill1.tga");
  for (i=0; i<10; i++) LandID[i] = TexID[GndID];
  land = new cLandscape[2];
  for (i=0; i<1; i++) {
    land[i].mMap.mTex = LandID;
    land[i].mMap.mShift = 5;
    land[i].mMap.mHShift = 3;
    if (land[i].mMap.LoadTga(buf)) Fatal("Error Loading HMap!");
    a = xmin(land[i].mMap.mWidth, land[i].mMap.mHeight);
    b = a / 2;
    land[i].mMap.CastShadows();
//    land[i].mPos.Set(-(land[i].mMap.mWidth/2)<<land[i].mMap.mShift, 0, -(land[i].mMap.mHeight/2)<<land[i].mMap.mShift);
//land[i].DrawList();
    planet->AddObject(cWorld::GROUND, &land[i]);
  };

  sprintf(buf, "%s%s", DATA_DIRECTORY, "textures/hill0.tga");
  for (i=0; i<10; i++) WaterID[i] = TexID[MscID+4];
  i = 1;
  land[i].mMap.mTex = WaterID;
  land[i].mMap.mShift = 5;
  land[i].mMap.mHShift = 3;
//  if (land[i].mMap.LoadTga(buf)) Fatal("Error Loading HMap!");
  land[i].mMap.Resize(land[0].mMap.mWidth, land[0].mMap.mHeight, 3);
  land[i].mMap.Clear(0, 96);
  land[i].mMap.Clear(1, 96);
  land[i].mMap.Clear(2, 255);
///  land[i].mPos.Set(-(land[i].mMap.mWidth/2)<<land[i].mMap.mShift, 0, -(land[i].mMap.mHeight/2)<<land[i].mMap.mShift);
  planet->AddObject(cWorld::WATER, &land[i]);

  a = a << land[0].mMap.mShift;
  SCALAR h;
  tree = new cTree[1024];
  for (i=0; i<1024; i++) {
    tree[i].mTex = TexID[TreID+rand()%4];
    tree[i].mShape.mWidth = 3+(rand()%2);
    tree[i].mShape.mHeight = 5+(rand()%3);
    tree[i].mShape.mAlignTo = 0 + 2 + 0; // Y-Axis only
    tree[i].mRot.Set(180, 0, 0);
    do {
      t.Set(rand()%a, tree[i].mShape.mHeight/2, rand()%a);
    } while((h = planet->GetHeight(t)) < 8);
    tree[i].mPos.Set(t.mV[X], t.mV[Y]+h, t.mV[Z]);
    tree[i].mOPos = tree[i].mPos;
    planet->AddObject(cWorld::NATURAL, &tree[i]);
  };


  sprintf(buf, "%s%s", DATA_DIRECTORY, "sprites/");

  weapon0 = new cWeapon;
  weapon0->mAmmo = new cProjectile;
  weapon0->mAmmo->mSpeed = 16;
  weapon0->mAmmo->mDamage = 16;
  weapon0->mAmmo->mTex = TexID[MscID+2];
  weapon0->mAmmo->mShape.mWidth = 2;
  weapon0->mAmmo->mShape.mHeight = 2;
  weapon0->mAmmo->mShape.mAlignTo = 1 + 2 + 4;
  weapon0->mOPos = weapon0->mPos;
  planet->AddObject(cWorld::WEAPON, weapon0);

  weapon1 = new cWeapon;
  weapon1->mAmmo = new cProjectile;
  weapon1->mAmmo->mSpeed = 16;
  weapon1->mAmmo->mDamage = 8;
  weapon1->mAmmo->mTex = TexID[SkyID+0];
  weapon1->mAmmo->mShape.mWidth = 1.5;
  weapon1->mAmmo->mShape.mHeight = 1.5;
  weapon1->mAmmo->mShape.mAlignTo = 0 + 0 + 0;
  weapon1->mOPos = weapon1->mPos;
  planet->AddObject(cWorld::WEAPON, weapon1);

  mech0 = new cMech;
  planet->AddObject(cWorld::MECHANICAL, mech0);
  mech0->Load(buf, "sprite0.lst");
  mech0->mSprite.AlignFace();
  mech0->mOPos = mech0->mPos;
  mech0->mWeapon = weapon0;
  mech0->mAnime = 1;

  mech1 = new cMech;
  planet->AddObject(cWorld::MECHANICAL, mech1);
  mech1->Load(buf, "sprite1.lst");
  mech1->mSprite.AlignFace();
  mech1->mPos.Set(5, 0, 5);
  mech1->mOPos = mech1->mPos;
  mech1->mWeapon = weapon1;
  mech1->mAnime = 0;
}


void Reshape(int w, int h)
{
  width = w;
  height = h;

#ifdef USE_3DFX
  if (width  > MAX_3DFX_XRES) width  = MAX_3DFX_XRES;
  if (height > MAX_3DFX_YRES) height = MAX_3DFX_YRES;
#endif

  mainView->mType = 0;
  mainView->Move(0, 0);
  mainView->Resize(width, height);

  glutPostRedisplay();
}


void Draw()
{
  if (paused) {
    usleep(USLEEP_DURING_PAUSE); // don't waste other programs time when paused!
    return;
  };

  usleep(USLEEP_PER_FRAME_WASTE); // ...waste some fps ;)

  cMatrix b, a, c, d;

  INTEG i, j;
  SCALAR h;

// Read in from /dev/js0 and /dev/js1 for joystick input
//

  if (!js_disabled) {  
    js0status = read(js0file, &js0, JS_RETURN);
    if (js0status != JS_RETURN) Fatal("/dev/js0 Read Error.\n");
    js1status = read(js1file, &js1, JS_RETURN);
    if (js1status != JS_RETURN) Fatal("/dev/js1 Read Error.\n");
  };

// Setup Planets Time (to determine SkyDome colors)
//

  planet->mDome->SetSystemDayTime();      // local time of "real" world
  planet->mDome->SetDayTime(12.00);     // High Noon!!!
//  planet->mDome->SetDayTime(24.00);     // Mid Night!!!
//  planet->mDome->SetDayTime(22.00);

// Setup to prepare to Draw
//

  cDome *dome = planet->mDome;

//  if (!wireframe) {
    glClearColor(dome->mLoColor.mV[X], dome->mLoColor.mV[Y], dome->mLoColor.mV[Z], 1.0);
    glFogfv(GL_FOG_COLOR, dome->mLoColor.mV);
/*
  } else {
    cVector dark;
    dark.Set(0.0, 0.0, 0.0, 1.0);
    glClearColor(0.0, 0.0, 0.0, 1.0);
    glFogfv(GL_FOG_COLOR, dark.mV);
  };
*/
  glEnable(GL_FOG);

  mainCamera->mRot.mV[X] -= yrSpeed * (mainFramer->mSPF * 35.0);
  mainCamera->mRot.mV[Y] -= xrSpeed * (mainFramer->mSPF * 35.0);
  SCALAR sp = 1.75 * (mainFramer->mSPF * 35.0) * (lButton-rButton);
  SCALAR cx = cos(mainCamera->mRot.mV[X]*0.017454);

///*
  mainCamera->mPos.mV[X] -= cx * sin(mainCamera->mRot.mV[Y]*0.017454) * sp;
  mainCamera->mPos.mV[Y] +=      sin(mainCamera->mRot.mV[X]*0.017454) * sp;
  mainCamera->mPos.mV[Z] -= cx * cos(mainCamera->mRot.mV[Y]*0.017454) * sp;
//*/

  sp = (0.75*5) * (lButton-rButton);

//  mech0->mRot = mainCamera->mRot;
  mech0->mRot.mV[X] = 0;
//  mech0->mGndThrust.mV[X] = -cx * sin(mainCamera->mRot.mV[Y]*0.017454) * sp;
//  mech0->mGndThrust.mV[Y] =       sin(mainCamera->mRot.mV[X]*0.017454) * sp;
//  mech0->mGndThrust.mV[Z] = -cx * cos(mainCamera->mRot.mV[Y]*0.017454) * sp;

//  mainCamera->mPos = mech0->mPos;
//  mainCamera->mPos.mV[Y] += 3.0;

//  mech1->mPos.mV[X] += sin(mech0->mRot.mV[Y]*0.017454) * 0.2 * (mainFramer->mSPF * 35.0);
//  mech1->mPos.mV[Z] += cos(mech0->mRot.mV[Y]*0.017454) * 0.2 * (mainFramer->mSPF * 35.0);
  mech1->Pattern(planet->mSeconds, "1ll1ll1ll1ll1nn1nn1rr1rr1rr1rr1rr1rr1rr1nn1nn");

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);  

  mainView->Use();

  a = mainCamera->GetCameraMatrix();  

if (kb0.buttons & 1) {
  weapon0->Fire();
  weapon1->Fire();
  kb0.buttons &= !kb0.buttons;
};

// ------------------------->
// - D R A W  -  B E L O W ->
// ------------------------->


// Do some To-World-Transformations and other stuff
//

  planet->Animate(mainFramer->mSPF);

/*
  mainCamera->mPos = mech0->mPos;
  mainCamera->mPos.mV[Y] += 3.0;
  mainCamera->mPos.mV[X] -= sin(mech0->mRot.mV[Y]*0.017454) * 10;  
  mainCamera->mPos.mV[Z] -= cos(mech0->mRot.mV[Y]*0.017454) * 10;
  a = mainCamera->GetCameraMatrix();
*/

  planet->ToWorld();

// Setup camera
//

  glLoadMatrixf(a.mM); // Use it!

// Setup the drawing scheme
//

  if (!wireframe) {
    glEnable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  } else {
    glDisable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  };

  glEnable(GL_CULL_FACE);

// Draw the environment: Ground, SkyDome, Clouds
//

  glDepthMask(GL_FALSE);  

//  dome->DrawGround(a, TexID[GndID]);

  if (!wireframe) {
    glDisable(GL_FOG);
    glDisable(GL_TEXTURE_2D);
    dome->DrawDome(a);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_FOG);
    dome->DrawClouds(a, TexID[SkyID+1]);
  };

  glDepthMask(GL_TRUE);  

  glDisable(GL_CULL_FACE);

// Draw the 3d Consoles  
//  
  
  b.Set();  
  b.MovRel(0, 0, -1);  
  if (console0) con[0]->Draw(b);  
  if (console1) con[1]->Draw(b);  


  glEnable(GL_CULL_FACE);


// Draw the things: Mechs, Buildings, Trees, Shots, ...
//

  cVector face = mainCamera->mRot;

  glColor3f(1.0, 1.0, 1.0);
  planet->Draw(a, face);


  glDisable(GL_CULL_FACE);

// Make the drawn things visible and setup some counters
//

  glFlush();
  glutSwapBuffers();

  mainFramer->Update(); // calculate frame time
}


void Idle()
{
  glutPostRedisplay();
}


void MouseFunc(int button, int state, int x, int y)
{
  if (button == GLUT_RIGHT_BUTTON)   rButton = state == GLUT_DOWN;
  if (button == GLUT_MIDDLE_BUTTON)  mButton = state == GLUT_DOWN;
  if (button == GLUT_LEFT_BUTTON)    lButton = state == GLUT_DOWN;
  if (x > width) x = width;
  if (y > height) y = height;
  mxpos = x;
  mypos = y;
  yrSpeed = (y-height/2.)/(height/4.);
  xrSpeed = (x-width/2.)/(width/4.);
  ms0.buttons = lButton;
  ms1.buttons = rButton;
}


void Mouse(int x, int y)
{
  if (x > width) x = width;
  if (y > height) y = height;
  mxpos = x;
  mypos = y;
  yrSpeed = (y-height/2.)/(height/4.);
  xrSpeed = (x-width/2.)/(width/4.);
}


void Key(unsigned char k, int x, int y)
{
  // maybe switch() would be better...

#ifdef USE_XMESA
  if (k == XMODE_KEY) {
    fullscreen = 1 - fullscreen;
    XMesaSetFXmode(fullscreen ? XMESA_FX_FULLSCREEN : XMESA_FX_WINDOW);
    if (fullscreen == 0) {
      cDummy->Insert("\nWINDOWED X MODE\n");
    } else {
      cDummy->Insert("\nFULLSCREEN X MODE\n");
    };
    if (paused) {
      paused = 0;
      Draw();
      paused = 1;
    };
  } else
#endif

  if (k == 27) {
#ifdef USE_XMESA
    XMesaSetFXmode(XMESA_FX_WINDOW);
#endif
    exit(0);
  } else
  if (k == '1') {
    kb1.y = 0;
  } else if (k == '2') {
    kb1.y = -200;
  } else if (k == '3') {
    kb1.y = -300;
  } else if (k == '4') {
    kb1.y = -400;
  } else if (k == '5') {
    kb1.y = -500;
  } else if (k == '6') {
    kb1.y = -600;
  } else if (k == '7') {
    kb1.y = -700;
  } else if (k == '8') {
    kb1.y = -800;
  } else if (k == '9') {
    kb1.y = -900;
  } else if (k == '0') {
    kb1.y = -1000;
  } else if (k == CAMERA_KEY) {
    external_camera = !external_camera;
  } else if (k == FIRE_KEY) {
    kb0.buttons |= 1;
  } else if (k == JUMPJET_KEY) {
    kb1.buttons |= 1;
  } else if (k == WIREFRAME_KEY) {
    wireframe = !wireframe;
  } else if (k == CONSOLE0_KEY) {
    console0 = !console0;
  } else if (k == CONSOLE1_KEY) {
    console1 = !console1;
  } else if (k == PAUSE_KEY) {
    sound->InitSound(SndID[0], 0, 1, 1.0, 1.0);
    if (!paused) {
      cDummy->Insert("\nPAUSED\n");
      Draw();
    } else {
      cDummy->Insert("\nSUSPEND\n");
    };
    paused = 1 - paused;
  } else {
    cDummy->InsertChar(k);
  };
}


void Special(int k, int x, int y)
{
  kb0.x = 0;
  if (k == GLUT_KEY_LEFT) kb0.x = -1000;
  if (k == GLUT_KEY_RIGHT) kb0.x = +1000;
  kb0.y = 0;
  if (k == GLUT_KEY_UP) kb0.y = -1000;
  if (k == GLUT_KEY_DOWN) kb0.y = +1000;
}



//
// GOD begins below!
//


#ifdef __win32

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
  int    argc = 2;
  char  *argv[] = { "Nothing", "Nothing", "Nothing" };

#else

int main(int argc, char **argv) {

#endif


  errlog = fopen("err.log", "w");

  char s[256];
  sound = new cSound;
  sound->InitSystem("127.0.0.1", 9433);
//  sound->SetRate(8000);

  if (!js_disabled) {
    js0file = open(js0name, O_RDONLY);
    if (js0file < 0) js_disabled = 1;
    js1file = open(js1name, O_RDONLY);
    if (js1file < 0) js_disabled = 1;
    js0status = read(js0file, &js0, JS_RETURN);
    if (js0status != JS_RETURN) js_disabled = 1;
    js1status = read(js1file, &js1, JS_RETURN);
    if (js1status != JS_RETURN) js_disabled = 1;
  };
  
  js0xm = js0.x; // joystick middle is set to the current joystick position
  js0ym = js0.y; // ie. autocalibration
  js1xm = js1.x;
  js1ym = js1.y;
  
  kb0.x = 0; // keyboard
  kb0.y = 0;
  kb1.x = 0;
  kb1.y = 0;

  ms0.x = 0; // mouse
  ms0.y = 0;
  ms1.x = 0;
  ms1.y = 0;

  width = 640;
  height = 480;

#ifdef USE_3DFX
  putenv("MESA_GLX_FX=f");
  putenv("FX_GLIDE_NO_SPLASH=");
  width   = MAX_3DFX_XRES;
  height  = MAX_3DFX_YRES;
#endif

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);
  glutInitWindowSize(width, height);
  glutInitWindowPosition(0, 0);
  glutCreateWindow("My Window");

  sprintf(s, "%s%s", DATA_DIRECTORY, "/sounds/warpout.wav");
  SndID[0] = sound->RegisterSound(s, 0);

  sprintf(s, "%s%s", DATA_DIRECTORY, "/sounds/func.wav");
  SndID[1] = sound->RegisterSound(s, 0);

  sound->InitSound(SndID[1], 0, 1, 1.0, 1.0);

#ifdef START_WITH_XFULLSCREEN
  glutFullScreen();
  fullscreen = 1 - fullscreen;
#endif

#ifdef USE_3DFX
//  glutFullScreen();
//  fullscreen = 1 - fullscreen;
  XMesaSetFXmode(fullscreen ? XMESA_FX_FULLSCREEN : XMESA_FX_WINDOW);
#endif

  glutDisplayFunc(*Draw);
  glutReshapeFunc(*Reshape);
  glutIdleFunc(*Idle);
  glutMouseFunc(*MouseFunc);
  glutPassiveMotionFunc(*Mouse);
  glutMotionFunc(*Mouse);
  glutKeyboardFunc(*Key);
  glutSpecialFunc(*Special);

  Init();
  glutMainLoop();

  if (!js_disabled) {  
    close(js1file);
    close(js0file);
  };

  fclose(errlog);

  return 0;
}


