#include "objects.h"



#ifndef M_PI
#define M_PI 3.141594f
#endif



//--------------------------------


cWorld::cWorld()
{
  mWind.Set(0.0, 0.0, 0.0);
  mGravity.Set(0.0, -0.98, 0.0);
  mAthmosphere = 0.1;
  mSeconds = 0;
  mP = 0;
  INTEG i, j;
  for (j=0; j<MAX_TYPES; j++) {
    mN[j] = 0;
    for (i=0; i<mMAX_OBJECTS; i++) {
      mO[j][i] = NULL;
    };
  };
}


SCALAR cWorld::GetHeight(cVector pos)
{
  INTEG i, j, x, z;
  cVector p;
  SCALAR h;
  cHBmp *map;
  j = GROUND;
  for (i=0; i<mN[j]; i++) {
    map = &((cLandscape*) mO[j][i])->mMap;
    p.Set(pos.mV[X]-mO[j][i]->mPos.mV[X], 0, pos.mV[Z]-mO[j][i]->mPos.mV[Z]);
    x = INTEG(p.mV[X])>>map->mShift;
    z = INTEG(p.mV[Z])>>map->mShift;
    if (inarea(x, 0, map->mWidth-1) && inarea(z, 0, map->mHeight-1)) {
      h = map->GetHeight(p);
      return ((h<0)?(0):(h));
    };
  };
  return 0;
}


void cWorld::AddObject(INTEG n, cObject *object)
{
  INTEG m = mN[n] % mMAX_OBJECTS;
  if (mO[n][m] != NULL) delete mO[n][m];
  mO[n][m] = object;
  mO[n][m]->mWorld = this;
  mN[n] = (mN[n] + 1);// % mMAX_OBJECTS;
}


void cWorld::Animate(SCALAR delta)
{
  INTEG i, j, n;
  mSeconds += delta;
  for (j=0; j<MAX_TYPES; j++) {
    n = mN[j];
    if (n >= mMAX_OBJECTS) n = mMAX_OBJECTS;
    for (i=0; i<n; i++) {
      if (mO[j][i] != NULL) mO[j][i]->Animate(delta);
    };
  };
}


void cWorld::ToWorld()
{
  INTEG i, j, n;
  for (j=0; j<MAX_TYPES; j++) {
    n = mN[j];
    if (n >= mMAX_OBJECTS) n = mMAX_OBJECTS;
    for (i=0; i<n; i++) {
      if (mO[j][i] == NULL) continue;
      mO[j][i]->ToWorld();
    };
  };
}


void cWorld::Draw(cMatrix c, cVector face)
{
  INTEG i, j, n;
  cVector p;
  cVector v;
  cMatrix inv;
  inv = c;
  inv.Inv();
  inv.Apply(p);
  for (j=0; j<MAX_TYPES; j++) {
    n = mN[j];
    if (n >= mMAX_OBJECTS) n = mMAX_OBJECTS;
    for (i=0; i<n; i++) {
      if (mO[j][i] == NULL) continue;
      v = mO[j][i]->mPos;
      v.Sub(p);
      if ((j != WATER) && (j != GROUND)) {
        if (xabs(v.mV[X]) + xabs(v.mV[Z]) > 512) continue;
      };
      mO[j][i]->Draw(c, face);
    };
  };
}



//--------------------------------



cObject::cObject()
{
  mWorld        = NULL;
  mName         = NULL;
}


INTEG cObject::Hit(cSphere sphere)
{
  return 0;
}


BOOLEAN cObject::Hit(cRay ray, cVector &point)
{
  return 0;
}


void cObject::ChasisLR(SCALAR r)
{
  SCALAR e = r;
  if (e>+1) e = +1;
  if (e<-1) e = -1;
  mRot.mV[Y] += e;
  xmod(mRot.mV[Y], 360);
}


void cObject::ChasisUD(SCALAR r)
{
  SCALAR e = r;
  if (e>+1) e = +1;
  if (e<-1) e = -1;
  mRot.mV[X] += e;
  xmod(mRot.mV[X], 360);
}


void cObject::TowerLR(SCALAR r)
{
  SCALAR e = r;
  if (e>+2) e = +2;
  if (e<-2) e = -2;
  mTwr.mV[Y] += e;
  if (mTwr.mV[Y] > +75) mTwr.mV[Y] = +75;
  if (mTwr.mV[Y] < -75) mTwr.mV[Y] = -75;
}


void cObject::TowerUD(SCALAR r)
{
  SCALAR e = r;
  if (e>+2) e = +2;
  if (e<-2) e = -2;
  mTwr.mV[X] += e;
  if (mTwr.mV[X] > +75) mTwr.mV[X] = +75;
  if (mTwr.mV[X] < -75) mTwr.mV[X] = -75;
}


void cObject::GndSpeed(cVector delta)
{
//  mGndThrust.Add(delta);
mGndThrust = delta;
  mGndThrust.mV[X] = fitarea(mGndThrust.mV[X], -100, +100);
  mGndThrust.mV[Y] = fitarea(mGndThrust.mV[Y], -100, +100);
  mGndThrust.mV[Z] = fitarea(mGndThrust.mV[Z], -100, +100);
}


void cObject::AirSpeed(cVector delta)
{
//  mAirThrust.Add(delta);
mAirThrust = delta;
  mAirThrust.mV[X] = fitarea(mAirThrust.mV[X], -100, +100);
  mAirThrust.mV[Y] = fitarea(mAirThrust.mV[Y], -100, +100);
  mAirThrust.mV[Z] = fitarea(mAirThrust.mV[Z], -100, +100);
}


void cObject::Pattern(SCALAR t, char *pattern)  
{  
  cVector vzero, vone;
//  vone.Set(0, 0, 0.1);
  INTEG l = strlen(pattern);  
  INTEG s = INTEG(t*2.0) % l;  
  char c = pattern[s];  
  switch (c) {  
    case 'l': ChasisLR(+5);   break;  
    case 'r': ChasisLR(-5);   break;  
    case '0': GndSpeed(vzero);  break;  
    case '1': GndSpeed(vone);  break;
    case 'n': break;  
    default:  break;  
  };  
}  


void cObject::Move(SCALAR delta, cVector grav, SCALAR gndfactor, SCALAR atmfactor)
{
  cVector e, g, h;
  mSpeed = mPos;
  mSpeed.Sub(mOPos);          // There is a small inaccuracy in speed which is the difference
  mOPos = mPos;               // between the time the second last frame took and the time the last
                              // frame took...if you understand...(last_delta-delta) anyway it
                              // shouldn't bother if the frame rate is relatively stable.
  g = mGndThrust;
  g.Mul(gndfactor);           // take the ground friction into acount
  g.Mul(delta);               // scale movement by the seconds per frame

  h = mAirThrust;
  h.Mul(atmfactor);           // take the athmosphere into acount
  h.Mul(delta);

  e = grav;
  e.Mul(delta);

  mSpeed.Add(g);
  mSpeed.Add(h);
//  mSpeed.Mul(1-gndfactor);
//  mSpeed.Mul(1-atmfactor);
  mSpeed.mV[X] *= 0.8;
  mSpeed.mV[Z] *= 0.8;
  mSpeed.Add(e);

  mPos.Add(mSpeed);
}


//--------------------------------


void cLandscape::Animate(SCALAR delta)
{
}


void cLandscape::ToWorld()  
{  
}  
  

void cLandscape::DrawList()  
{  
  cMatrix super;
  mGeoList = glGenLists(1);
  glNewList(mGeoList, GL_COMPILE);
  mMap.Draw(super);
  glEndList();
}  

  
void cLandscape::Draw(cMatrix c, cVector face)  
{  
  glColor3f(1,1,1);
  cMatrix super;
  super.RotXYZ(mRot.mV[X], mRot.mV[Y], mRot.mV[Z]);
  super.MovRel(mPos.mV[X], mPos.mV[Y], mPos.mV[Z]);
  super.Mul(c);

  cVector p, q;
  cMatrix inv;

  inv = super;
  inv.Inv();
  inv.Apply(p);

  q.Set(0, 0, 1);
  inv.Apply(q);

  mMap.Draw(super, 40, 40, 1, p.mV[X] / (1<<mMap.mShift), p.mV[Z] / (1<<mMap.mShift));
}  



//-----------------------------------



void cTree::Animate(SCALAR delta)
{
  cVector v;
  v = mGndThrust;
  v.Add(mAirThrust);
  v.Mul(delta);
  mOPos = mPos;
  mPos.Add(v);
}


void cTree::ToWorld()
{
}


void cTree::Draw(cMatrix c, cVector face)
{
  cVector f = face;
  f.mV[X] = 0;
  glColor3f(1,1,1);
  glBindTexture(GL_TEXTURE_2D, mTex);
  mShape.mPosition = mPos;
  mShape.mRotation = mRot;
  mShape.Draw(c, f);
}


//-----------------------------------  


void cMech::Animate(SCALAR delta)
{
  INTEG i, bone;
  for (i=0; i<mBody.mAnimations->mAnimations[mAnime].mNAniBones; i++) {
    if (mBody.mAnimations->mAnimations[mAnime].mAniBones[i].mBone >= 0) { // is valid bone?
      bone = mBody.mAnimations->mAnimations[mAnime].mAniBones[i].mBone;
      mBody.mBones[bone].Animate(
        delta,
        mBody.mAnimations->mAnimations[mAnime].mAniBones[i].mNKFrames,
        mBody.mAnimations->mAnimations[mAnime].mAniBones[i].mKFrames
      );
    };
  };
  cVector l, r, s;
  SCALAR hl, hr;
  l = mBody.mBones[mLFoot].mLJoint;
  r = mBody.mBones[mRFoot].mLJoint;
  l.Add(mPos);
  r.Add(mPos);
  hl = mWorld->GetHeight(l);
  hr = mWorld->GetHeight(r);
  l.Sub(mPos);
  r.Sub(mPos);
  if (hl-l.mV[Y] > hr-r.mV[Y]) {
  if (mPos.mV[Y] < hl-l.mV[Y]) {
    mPos.mV[Y] = hl-l.mV[Y];
    if (mOPos.mV[Y] < mPos.mV[Y]) mOPos.mV[Y] = mPos.mV[Y];
    s = mBody.mBones[mLFoot].mOLJoint;
    s.Sub(mBody.mBones[mLFoot].mLJoint);
    s.mV[Y] = 0;
    mPos.Add(s);
  };
  } else {
  if (mPos.mV[Y] < hr-r.mV[Y]) {
    mPos.mV[Y] = hr-r.mV[Y];
    if (mOPos.mV[Y] < mPos.mV[Y]) mOPos.mV[Y] = mPos.mV[Y];
    s = mBody.mBones[mRFoot].mOLJoint;
    s.Sub(mBody.mBones[mRFoot].mLJoint);
    s.mV[Y] = 0;
    mPos.Add(s);
  };
  };
  Move(delta, mWorld->mGravity, 0.6, mWorld->mAthmosphere);

  mWeapon->mPos = mBody.mBones[mRHand].mLJoint;
  mWeapon->mPos.Add(mPos);
  mWeapon->mRot = mRot;

  cMatrix mhand;
  cMatrix mzero;
  cVector vzero;
  mhand = mBody.mBones[mRHand].mMatrix;
}


void cMech::Load(char *prefix, char *fname)
{
  SCALAR e;
  char p[1024], pp[1024];
  sprintf(p, "%s%s", prefix, fname);
  FILE *f, *fx;
  INTEG i, n;
  char buf[81], b[82];;
  cBmp *bmp = new cBmp;

  f = fopen(p, "r"); 

  fscanf(f, "%80s\n", b);
  sprintf(pp, "%s%s", prefix, b);
report(pp);
report("\n");

  fscanf(f, "%80s\n", b);
  sprintf(buf, "%s%s", pp, b);
report(buf);
report("\n");
  bmp->LoadTga(buf);
  bmp->glRegister(1, mSprite.mTexture[0]);

  fscanf(f, "%80s\n", b);
  sprintf(buf, "%s%s", pp, b);
report(buf);
report("\n");
  bmp->LoadTga(buf);
  bmp->glRegister(1, mSprite.mTexture[1]);

  fscanf(f, "%80s\n", b);
  sprintf(buf, "%s%s", pp, b);
report(buf);
report("\n");
  mSprite.GetMem(256, 512);
  mSprite.Load(buf);

  fscanf(f, "%80s\n", b);
  sprintf(buf, "%s%s", pp, b);
report(buf);
report("\n");
  fx = fopen(buf, "r");
  mBody.Load(fx);
  fclose(fx);

  fscanf(f, "%i\n", &n);
  mBody.mAnimations = new cAnime;
  mBody.mAnimations->mNAnimations = n;  
  mBody.mAnimations->mAnimations = new cAniBody[mBody.mAnimations->mNAnimations];
  for (i=0; i<n; i++)  
  {  
    fscanf(f, "%s\n", b);
    sprintf(buf, "%s%s", pp, b);
report(buf);
report("\n");
    fx = fopen(buf, "r");
    mBody.mAnimations->mAnimations[i].Load(fx);
    fclose(fx);
  };  

  fscanf(f, "%f\n", &e);
  mSprite.Scale(e);
  mBody.Scale(e);

  mBody.mAnimations->Assign(mBody.mNBones, mBody.mBones);
  mLFoot = mBody.mBones[0].FindBone("lfoot", mBody.mNBones, mBody.mBones);
  mRFoot = mBody.mBones[0].FindBone("rfoot", mBody.mNBones, mBody.mBones);
  mLHand = mBody.mBones[0].FindBone("lhand", mBody.mNBones, mBody.mBones);
  mRHand = mBody.mBones[0].FindBone("rhand", mBody.mNBones, mBody.mBones);

  fclose(f);

  cVector vzero;
  cMatrix mzero;
  mBody.mBones[0].BaseToWorld(mzero, vzero);
  mAnime = 0;
}


void cMech::ToWorld()
{
  cMatrix mzero;
  cVector vzero;
  mBody.mBones[0].ToWorld(mzero, mzero, vzero);
}


void cMech::Draw(cMatrix c, cVector face)
{
  glColor3f(1,1,1);
  cMatrix super;
  super.RotXYZ(mRot.mV[X], mRot.mV[Y], mRot.mV[Z]);
  super.MovRel(mPos.mV[X], mPos.mV[Y], mPos.mV[Z]);
  super.Mul(c);
//  mBody.mBones[1].Draw(super);
  mSprite.Draw(super, &mBody);
}



//-------------------------------------------



void cWeapon::Load(char *prefix, char *fname)
{
  char p[1024];
  sprintf(p, "%s%s", prefix, fname);
  FILE *f;
}


void cWeapon::Animate(SCALAR delta)
{
}


void cWeapon::Fire()
{
  if (mAmmo == NULL) return;
  cProjectile *p = new cProjectile;
  if (p != NULL) {
    *p = *mAmmo;
    cMatrix super;
    super.RotXYZ(mRot.mV[X], mRot.mV[Y], mRot.mV[Z]);
    p->mPos = mPos;
    p->mOPos = mPos;
    p->mAirThrust.Set(0, 0, p->mSpeed);
    super.Apply(p->mAirThrust);
/*
    p->mPos = mPos;
    p->mShape.mWidth = mAmmo->mShape.mWidth;
    p->mShape.mHeight = mAmmo->mShape.mHeight;
    p->mShape.mAlignTo = mAmmo->mShape.mAlignTo;
*/
    mWorld->AddObject(cWorld::PROJECTILE, p);
  };
}


void cWeapon::ToWorld()
{
}


void cWeapon::Draw(cMatrix c, cVector face)
{
  glColor3f(1,1,1);
  cMatrix super;
  super.RotXYZ(mRot.mV[X], mRot.mV[Y], mRot.mV[Z]);
  super.MovRel(mPos.mV[X], mPos.mV[Y], mPos.mV[Z]);
  super.Mul(c);
}



//-------------------------------------------



void cProjectile::Draw(cMatrix c, cVector face)
{
  glColor3f(1,1,1);
  glBindTexture(GL_TEXTURE_2D, mTex);
  mShape.mPosition = mPos;
  mShape.mRotation = mRot;
  mShape.Draw(c, face);
}



