#ifndef OBJECTS_INC
#define OBJECTS_INC

#include <psi3d/psi3d.h>

//-----------------------------------

class cWorld;

class cObject;
  class cLandscape;
  class cTree;
  class cMech;
  class cWeapon;
  class cProjectile;


//-----------------------------------


class cWorld {
  public:
  const static INTEG mMAX_OBJECTS    = 1024;
  public:
  enum             { MILITARY, CIVILISTIC, MAX_GROUPS };
  enum             { NEUTRAL_ALIANCE, A_ALLIANCE, B_ALLIANCE, C_ALLIANCE, D_ALLIANCE, MAX_ALLIANCES };
  enum             { GROUND, MECHANICAL, WATER, WEAPON, NATURAL, PROJECTILE, MAX_TYPES };
  public:
  cDome             *mDome;
  INTEG              mP;
  INTEG              mN[MAX_TYPES];
  cObject           *mO[MAX_TYPES][mMAX_OBJECTS];
  SCALAR             mAthmosphere;
  cVector            mGravity;
  cVector            mWind;
  SCALAR             mSeconds;
  public:
                     cWorld();
  SCALAR             GetHeight(cVector pos);
  void               AddObject(INTEG n, cObject *object);
  void               Animate(SCALAR delta);
  void               ToWorld();
  void               Draw(cMatrix c, cVector face);
};


//-----------------------------------


class cObject {
  public:
  cWorld            *mWorld;
  INTEG              mType;
  INTEG              mGroup;
  INTEG              mAlliance;
  char              *mName;
  cVector            mOPos;
  cVector            mPos;
  cVector            mRot;
  cVector            mTwr;
  cVector            mGndThrust; // *caution*: these are additive (acceleration)
  cVector            mAirThrust;
  cVector            mSpeed;
  public:

                     cObject();
  virtual INTEG      Hit(cSphere sphere);
  virtual BOOLEAN    Hit(cRay ray, cVector &point);
  virtual void       ChasisLR(SCALAR r);
  virtual void       ChasisUD(SCALAR r);
  virtual void       TowerLR(SCALAR r);
  virtual void       TowerUD(SCALAR r);
  virtual void       GndSpeed(cVector delta);
  virtual void       AirSpeed(cVector delta);
  virtual void       Pattern(SCALAR t, char *pattern);
  virtual void       Move(SCALAR delta, cVector grav, SCALAR gndfactor, SCALAR atmfactor);
  virtual void       Animate(SCALAR delta) = 0;
  virtual void       ToWorld() = 0;

 virtual void       Draw(cMatrix c, cVector face) = 0;
};


class cLandscape : public cObject {
  public:
  GLuint             mGeoList; // Draw List
  cHBmp              mMap;  // the height (bit)map
  public:
  void               Animate(SCALAR delta);
  void               ToWorld();
  void               DrawList();
  void               Draw(cMatrix c, cVector face);
};


class cTree : public cObject {
  public:
  WORD               mTex;    // texture to use
  cParticle          mShape;  // for sprite drawing
  public:
  void               Animate(SCALAR delta);
  void               ToWorld();
  void               Draw(cMatrix c, cVector face);
};


class cMech : public cObject {
  public:
  cSprite            mSprite;  // skin &
  cBody              mBody;    // bones
  INTEG              mAnime;   // current running animation
  INTEG              mLFoot;
  INTEG              mRFoot;
  INTEG              mLHand;
  INTEG              mRHand;
  cWeapon           *mWeapon;
  public:
  void               Load(char *prefix, char *fname);
  void               Animate(SCALAR delta);
  void               ToWorld();
  void               Draw(cMatrix c, cVector face);
};


class cWeapon : public cObject {
  public:
  cProjectile       *mAmmo;
  public:
  void               Load(char *prefix, char *fname);
  void               Animate(SCALAR delta);
  void               Fire();
  void               ToWorld();
  void               Draw(cMatrix c, cVector face);
};


class cProjectile : public cTree { // The inheritance sux someway, but...
  public:
  INTEG              mDamage;
  SCALAR             mSpeed;
  public:
  void               Draw(cMatrix c, cVector face);
};


#endif
